import contextlib
from typing import Callable, Optional, Union

from twisted.internet import reactor
from web3 import Web3
from web3.exceptions import TransactionNotFound
from web3.types import PendingTx as PendingTxData, TxData, TxParams
from web3.types import RPCError, TxReceipt, Wei

from atxm.exceptions import (
    InsufficientFunds,
)
from atxm.logging import log
from atxm.tx import AsyncTx, FutureTx


def _get_average_blocktime(w3: Web3, sample_size: int) -> float:
    """Returns the average block time in seconds."""
    latest_block = w3.eth.get_block("latest")
    if latest_block.number == 0:
        return 0
    sample_block_number = latest_block.number - sample_size
    if sample_block_number <= 0:
        return 0
    base_block = w3.eth.get_block(sample_block_number)
    delta = latest_block.timestamp - base_block.timestamp
    average_block_time = delta / sample_size
    return average_block_time


def _log_gas_weather(base_fee: Wei, tip: Wei) -> None:
    base_fee_gwei = Web3.from_wei(base_fee, "gwei")
    tip_gwei = Web3.from_wei(tip, "gwei")
    log.info(f"Gas conditions: base {base_fee_gwei} gwei | tip {tip_gwei} gwei")


def _get_receipt(w3: Web3, data: Union[TxData, PendingTxData]) -> Optional[TxReceipt]:
    try:
        receipt = w3.eth.get_transaction_receipt(data["hash"])
    except TransactionNotFound:
        return
    return receipt


def fire_hook(hook: Callable, tx: AsyncTx, *args, **kwargs) -> None:
    """
    Fire a callable in a separate thread.
    Try exceptionally hard not to crash the async tasks during dispatch.
    """
    with contextlib.suppress(Exception):

        def _hook() -> None:
            """I'm inside a thread!"""
            try:
                hook(tx, *args, **kwargs)
            except Exception as e:
                log.warn(f"[hook] {e}")

        reactor.callInThread(_hook)
        log.info(f"[hook] fired hook {hook} for transaction #atx-{tx.id}")


def _handle_rpc_error(e: Exception, tx: FutureTx) -> None:
    try:
        error = RPCError(**e.args[0])
    except TypeError:
        log.critical(
            f"[error] transaction #atx-{tx.id}|{tx.params['nonce']} failed with {e}"
        )
    else:
        log.critical(
            f"[error] transaction #atx-{tx.id}|{tx.params['nonce']} failed with {error['code']} | {error['message']}"
        )
        if error["code"] == -32000:
            if "insufficient funds" in error["message"]:
                raise InsufficientFunds
        hook = tx.on_fault
        if hook:
            fire_hook(hook=hook, tx=tx, error=e)


def _make_tx_params(data: TxData) -> TxParams:
    """
    TxData -> TxParams: Creates a transaction parameters
    object from a transaction data object for broadcast.

    This operation is performed in order to "turnaround" the transaction
    data object as queried from the RPC provider (eth_getTransaction) into a transaction
    parameters object for strategics and re-broadcast (LocalAccount.sign_transaction).
    """
    params = TxParams(
        {
            "nonce": data["nonce"],
            "chainId": data["chainId"],
            "gas": data["gas"],
            "to": data["to"],
            "value": data["value"],
            "data": data.get("data", b""),
        }
    )
    if "gasPrice" in data:
        params["type"] = "0x01"
        params["gasPrice"] = data["gasPrice"]
    elif "maxFeePerGas" in data:
        params["type"] = "0x02"
        params["maxFeePerGas"] = data["maxFeePerGas"]
        params["maxPriorityFeePerGas"] = data["maxPriorityFeePerGas"]
    return params



from copy import deepcopy
from logging import Logger
from typing import Optional, Union, List, Type

from eth_account.signers.local import LocalAccount
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.task import LoopingCall
from web3 import Web3
from web3.exceptions import TransactionNotFound
from web3.types import TxReceipt, TxParams

from atxm.exceptions import (
    Wait,
    TransactionReverted,
    Fault,
    Faults,
)
from atxm.state import _State
from atxm.strategies import (
    AsyncTxStrategy,
    InsufficientFundsPause,
    TimeoutPause,
    FixedRateSpeedUp,
)
from atxm.tx import (
    FinalizedTx,
    FutureTx,
    PendingTx,
    TxHash,
)
from atxm.utils import (
    _get_average_blocktime,
    _get_receipt,
    _handle_rpc_error,
    fire_hook,
    _make_tx_params,
)
from .logging import log


class _Machine:
    """
    Do not import this - it will not work.
    Please import the publicly exposed `AutomaticTxMachine` instead.
    Thanks! :)
    """

    # internal
    __CLOCK = reactor  # twisted reactor

    # tweaks
    _TRACKING_CONFIRMATIONS = 300  # blocks until clearing a finalized transaction
    _RPC_THROTTLE = 1  # min. seconds between RPC calls (>1 recommended)
    _MIN_INTERVAL = 1
    _IDLE_INTERVAL = 60 * 5  # seconds
    _BLOCK_INTERVAL = 20  # ~20 blocks
    _BLOCK_SAMPLE_SIZE = 10_000  # blocks

    STRATEGIES: List[Type[AsyncTxStrategy]] = [
        InsufficientFundsPause,
        TimeoutPause,
        FixedRateSpeedUp,
    ]

    def __init__(
        self,
        w3: Web3,
        strategies: Optional[List[AsyncTxStrategy]] = None,
        disk_cache: bool = False,
    ):
        # public
        self.w3 = w3
        self.signers = {}
        self.log = Logger(self.__class__.__name__)
        self.strategies = [s(w3) for s in self.STRATEGIES]
        if strategies:
            self.strategies.extend(list(strategies))

        # state
        self.__pause = False
        self._state = _State(disk_cache=disk_cache)

        # async
        self._task = LoopingCall(self._cycle)
        self._task.clock = self.__CLOCK
        self._task.interval = self._IDLE_INTERVAL

    @property
    def _clock(self):
        return self.__CLOCK

    @property
    def _interval(self) -> Optional[float]:
        return self._task.interval

    @property
    def _start_time(self) -> float:
        return self._task.starttime

    @property
    def _busy(self) -> bool:
        """Returns True if the machine is busy."""
        return bool(self._state.queue or self._state.pending)

    #
    # Async
    #

    def _handle_errors(self, *args, **kwargs):
        """Handles unexpected errors during task processing."""
        self._state.commit()
        self.log.warn(
            "[recovery] error during transaction: {}".format(args[0].getTraceback())
        )
        if self._task.running:
            return
        self.log.warn("[recovery] restarting transaction machine!")
        self._start(now=False)

    def _cycle(self) -> None:
        """Execute one cycle"""

        if self.__pause:
            self.log.warn("[pause] paused")
            return

        self.__monitor_finalized()
        if not self._busy:
            self.log.info(f"[idle] cycle interval is {self._task.interval} seconds")
            return

        # sample block time and adjust the interval
        # every work cycle.
        self.__work_mode()
        self.log.info(
            f"[working] tracking {len(self._state.queue)} queued "
            f"transaction{'s' if len(self._state.queue) > 1 else ''} "
            f"{'and 1 pending transaction' if self._state.pending else ''}"
        )

        if self._state.pending:
            # There is an active transaction
            self.__handle_active_transaction()

        elif self._state.queue:
            # There is no active transaction
            # and there are queued transactions
            self.__broadcast()

        # After one work cycle, return to idle mode
        # if the machine is not busy.
        if not self._busy:
            self.__idle_mode()

    #
    # Throttle
    #

    def _start(self, now: bool = False) -> Deferred:
        if self._task.running:
            return self._task.deferred
        when = "now" if now else f"in {self._task.interval} seconds"
        self.log.info(f"[atxm] starting async transaction machine {when}")
        d = self._task.start(interval=self._interval, now=now)
        d.addErrback(self._handle_errors)
        return d

    def _stop(self):
        """Stop task."""
        if self._task.running:
            self._task.stop()

    def __idle_mode(self) -> None:
        """Return to idle mode (slow down)"""
        self._task.interval = self._IDLE_INTERVAL
        self.log.info(
            f"[done] returning to idle mode with "
            f"{self._task.interval} second interval"
        )
        if not self._busy:
            self._sleep()

    def __work_mode(self) -> None:
        """Start work mode (speed up)"""
        average_block_time = _get_average_blocktime(
            w3=self.w3, sample_size=self._BLOCK_SAMPLE_SIZE
        )
        self._task.interval = max(
            round(average_block_time * self._BLOCK_INTERVAL), self._MIN_INTERVAL
        )
        self.log.info(f"[working] cycle interval is {self._task.interval} seconds")

    def _wake(self) -> None:
        if not self._task.running:
            log.info("[wake] waking up")
            self._start(now=True)

    def _sleep(self) -> None:
        if self._task.running:
            log.info("[sleep] sleeping")
            self._stop()

    #
    # Lifecycle
    #

    def __handle_active_transaction(self) -> bool:
        """
        Handles the currently tracked pending transaction.

        The 4 possible outcomes for the pending ("active") transaction in one cycle:

        1. paused
        2. reverted (fault)
        3. finalized
        4. strategize: retry, wait, or fault

        Returns True if the next queued transaction can be broadcasted right now.
        """

        # Outcome 1: the pending transaction is paused by strategies
        if self.__pause:
            self.log.warn(
                f"[pause] transaction #{self._state.pending.id} is paused by strategies"
            )
            return False

        try:
            receipt = self.__get_receipt()

        # Outcome 2: the pending transaction was reverted (final error)
        except TransactionReverted:
            self._state.fault(
                error=self._state.pending.txhash.hex(),
                fault=Faults.REVERT,
                clear_active=True,
            )
            return True

        # Outcome 3: pending transaction is finalized (final success)
        if receipt:
            final_txhash = receipt["transactionHash"]
            confirmations = self.__get_confirmations(tx=self._state.pending)
            self.log.info(
                f"[finalized] Transaction #atx-{self._state.pending.id} has been finalized "
                f"with {confirmations} confirmations txhash: {final_txhash.hex()}"
            )
            self._state.finalize_active_tx(receipt=receipt)
            return True

        # Outcome 4: re-strategize the pending transaction
        pending_tx = self.__strategize()
        return pending_tx is not None

    #
    # Broadcast
    #

    def __get_signer(self, address: str) -> LocalAccount:
        try:
            signer = self.signers[address]
        except KeyError:
            raise ValueError(f"Signer {address} not found")
        return signer

    def __fire(self, tx: FutureTx, msg: str) -> Optional[PendingTx]:
        """
        Signs and broadcasts a transaction, handling RPC errors
        and internal state changes.

        On success, returns the `PendingTx` object.
        On failure, returns None.

        Morphs a `FutureTx` into a `PendingTx` and advances it
        into the active transaction slot if broadcast is successful.
        """
        signer: LocalAccount = self.__get_signer(tx._from)
        try:
            txhash = self.w3.eth.send_raw_transaction(
                signer.sign_transaction(tx.params).rawTransaction
            )
        except ValueError as e:
            _handle_rpc_error(e, tx=tx)
            return
        self.log.info(
            f"[{msg}] fired transaction #atx-{tx.id}|{tx.params['nonce']}|{txhash.hex()}"
        )
        pending_tx = self._state.morph(tx=tx, txhash=txhash)
        if tx.on_broadcast:
            fire_hook(hook=tx.on_broadcast, tx=tx)
        return pending_tx

    def pause(self) -> None:
        self.__pause = True
        self.log.warn(
            f"[pause] pending transaction {self._state.pending.txhash.hex()} has been paused."
        )
        hook = self._state.pending.on_pause
        if hook:
            fire_hook(hook=hook, tx=self._state.pending)

    def resume(self) -> None:
        self.log.info("[pause] pause lifted by strategy")
        self.__pause = False  # resume

    def __strategize(self) -> Optional[PendingTx]:
        """Retry the currently tracked pending transaction with the configured strategy."""
        if not self._state.pending:
            raise RuntimeError("No active transaction to strategize")

        _active_copy = deepcopy(self._state.pending)
        for strategy in self.strategies:
            try:
                params = strategy.execute(pending=_active_copy)
            except Wait:
                self.pause()
                return
            except Fault as e:
                self._state.fault(
                    error=self._state.pending.txhash.hex(),
                    fault=e.fault,
                    clear_active=e.clear,
                )
                return
            if params:
                # in case the strategy accidentally returns None
                # keep the parameters as they are.
                _active_copy.params.update(params)

            if self.__pause:
                self.resume()

        # (!) retry the transaction with the new parameters
        retry_params = TxParams(_active_copy.params)
        _names = " -> ".join(s.name for s in self.strategies)
        pending_tx = self.__fire(tx=retry_params, msg=_names)
        self.log.info(f"[retry] transaction #{pending_tx.id} has been re-broadcasted")

        return pending_tx

    def __broadcast(self) -> Optional[TxHash]:
        """
        Attempts to broadcast the next `FutureTx` in the queue.
        If the broadcast is not successful, it is re-queued.
        """
        future_tx = self._state._pop()  # popleft
        future_tx.params = _make_tx_params(future_tx.params)
        signer = self.__get_signer(future_tx._from)
        nonce = self.w3.eth.get_transaction_count(signer.address, "latest")
        if nonce > future_tx.params["nonce"]:
            self.log.warn(
                f"[broadcast] nonce {future_tx.params['nonce']} has been front-run "
                f"by another transaction. Updating queued tx nonce {future_tx.params['nonce']} -> {nonce}"
            )
        future_tx.params["nonce"] = nonce
        pending_tx = self.__fire(tx=future_tx, msg="broadcast")
        if not pending_tx:
            self._state._requeue(future_tx)
            return
        return pending_tx.txhash

    #
    # Monitoring
    #

    def __get_receipt(self) -> Optional[TxReceipt]:
        """
        Hits eth_getTransaction and eth_getTransactionReceipt
        for the active pending txhash and checks if
        it has been finalized or reverted.

        Returns the receipt if the transaction has been finalized.
        NOTE: Performs state changes
        """
        try:
            txdata = self.w3.eth.get_transaction(self._state.pending.txhash)
            self._state.pending.data = txdata
        except TransactionNotFound:
            self.log.error(
                f"[error] Transaction {self._state.pending.txhash.hex()} not found"
            )
            return

        receipt = _get_receipt(w3=self.w3, data=txdata)
        status = receipt.get("status")
        if status == 0:
            # If status in response equals 1 the transaction was successful.
            # If it is equals 0 the transaction was reverted by EVM.
            # https://web3py.readthedocs.io/en/stable/web3.eth.html#web3.eth.Eth.get_transaction_receipt
            log.info(
                f"Transaction {txdata['hash'].hex()} was reverted by EVM with status {status}"
            )
            raise TransactionReverted(receipt)

        if receipt:
            log.info(
                f"[accepted] Transaction {txdata['nonce']}|{txdata['hash'].hex()} "
                f"has been included in block #{txdata['blockNumber']}"
            )
            return receipt

    def __get_confirmations(self, tx: Union[PendingTx, FinalizedTx]) -> int:
        current_block = self.w3.eth.block_number
        txhash = (
            tx.txhash if isinstance(tx, PendingTx) else tx.receipt["transactionHash"]
        )
        try:
            tx_receipt = self.w3.eth.get_transaction_receipt(txhash)
            tx_block = tx_receipt["blockNumber"]
            confirmations = current_block - tx_block
            return confirmations
        except TransactionNotFound:
            self.log.info(f"Transaction {txhash.hex()} is pending or unconfirmed")
            return 0

    def __monitor_finalized(self) -> None:
        """Follow-up on finalized transactions for a little while."""
        if not self._state.finalized:
            return
        for tx in self._state.finalized.copy():
            confirmations = self.__get_confirmations(tx=tx)
            txhash = tx.receipt["transactionHash"]
            if confirmations >= self._TRACKING_CONFIRMATIONS:
                if tx in self._state.finalized:
                    self._state.finalized.remove(tx)
                    self.log.info(
                        f"[clear] stopped tracking {txhash.hex()} after {confirmations} confirmations"
                    )
                continue
            self.log.info(
                f"[monitor] transaction {txhash.hex()} has {confirmations} confirmations"
            )



from enum import Enum

from web3.types import RPCError


class Faults(Enum):
    """
    Fault codes for transaction processing.
    These are alternate states that a transaction can enter
    other than "finalized".
    """

    # Strategy has been running for too long
    TIMEOUT = "timeout"

    # Transaction has been capped and subsequently timed out
    PAUSE = "pause"

    # Transaction reverted
    REVERT = "revert"

    # Something went wrong
    ERROR = "error"

    # ...
    INSUFFICIENT_FUNDS = "insufficient_funds"


class TransactionFinalized(Exception):
    """raised when a transaction has been included in a block"""


class InsufficientFunds(RPCError):
    """raised when a transaction exceeds the spending cap"""


class Wait(Exception):
    """
    Raised when a strategy exceeds a limitation.
    Used to mark a pending transaction as "wait, don't retry".
    """


class Fault(Exception):
    """raised when a transaction has been faulted"""

    def __init__(self, tx, fault: Faults, clear: bool, message: str):
        self.tx = tx
        self.fault = fault
        self.message = message
        self.clear = clear
        super().__init__(message)


class TransactionReverted(Exception):
    """raised when a transaction has been reverted"""



import time
from abc import ABC
from typing import Tuple, Optional

from web3 import Web3
from web3.types import Gwei, TxParams, Wei, PendingTx

from atxm.exceptions import (
    Wait,
    Fault,
    Faults,
)
from atxm.logging import log
from atxm.utils import (
    _log_gas_weather,
)


class AsyncTxStrategy(ABC):
    """Abstract base class for transaction strategies."""

    _NAME = NotImplemented

    def __init__(self, w3: Web3):
        self.w3 = w3
        self.log = log

    @property
    def name(self) -> str:
        """Used to identify the strategy in logs."""
        return self._NAME

    def execute(self, pending: PendingTx) -> TxParams:
        """
        Execute the strategy.

        Called by the transaction machine when a
        transaction is ready to be strategized. Accepts a PendingTx
        object with data from the most recent previous attempt
        (like tx.txhash, tx.params, tx.created, etc).

        This method must do one of the following:
        - Raise `Wait` to pause retries and wait around for a bit.
        - Raise `Fault`to signal the transaction cannot be retried.
        - Returns a TxParams dictionary to use in the next attempt.

        NOTE: Do not mutate the input `tx` object. Return a new TxParams
        dictionary with the updated transaction parameters. The input
        object is already deeply copied to avoid accidental mutation,
        but it's best to be mindful of this.

        CAUTION: please be mindful that the purpose of this middleware
        is to mutate transaction parameters and not to broadcast
        transactions. Broadcasting transactions is handled by the
        transaction machine and not by the strategies.

        WARNING: The parameters returned by this method will be
        signed by a hot wallet and broadcast to the network immediately.
        Please be mindful of the security implications of the
        parameters you return.

        """
        raise NotImplementedError


class InsufficientFundsPause(AsyncTxStrategy):
    """Pause strategy for pending transactions."""

    _NAME = "insufficient-funds"

    def execute(self, pending: PendingTx) -> TxParams:
        balance = self.w3.eth.get_balance(pending._from)
        if balance == 0:
            self.log.warn(
                f"Insufficient funds for transaction #{pending.params['nonce']}"
            )
            raise Fault(
                tx=pending,
                fault=Faults.INSUFFICIENT_FUNDS,
                message="Insufficient funds",
                clear=False,
            )
        # log.warn(f"Insufficient funds for transaction #{pending.params['nonce']}")
        # raise Wait("Insufficient funds")
        return pending.params


class TimeoutPause(AsyncTxStrategy):
    """Pause strategy for pending transactions."""

    _NAME = "timeout"
    _TIMEOUT = 60 * 60  # 1 hour in seconds

    def __init__(self, w3: Web3, timeout: Optional[int] = None):
        super().__init__(w3)
        self.timeout = timeout or self._TIMEOUT

    def __active_timed_out(self, pending: PendingTx) -> bool:
        """Returns True if the active transaction has timed out."""
        if not pending:
            return False
        timeout = (time.time() - pending.created) > self.timeout
        if timeout:
            return True

        time_remaining = round(self.timeout - (time.time() - pending.created))
        minutes = round(time_remaining / 60)
        remainder_seconds = time_remaining % 60
        end_time = time.time() + time_remaining
        human_end_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end_time))
        if time_remaining < (60 * 2):
            self.log.warn(
                f"[timeout] Transaction {pending.txhash.hex()} will timeout in "
                f"{minutes}m{remainder_seconds}s at {human_end_time}"
            )
        else:
            self.log.info(
                f"[pending] {pending.txhash.hex()} \n"
                f"[pending] {round(time.time() - pending.created)}s Elapsed | "
                f"{minutes}m{remainder_seconds}s Remaining | "
                f"Timeout at {human_end_time}"
            )
        return False

    def execute(self, pending: PendingTx) -> TxParams:
        timeout = self.__active_timed_out(pending)
        if timeout:
            raise Fault(
                tx=pending,
                fault=Faults.TIMEOUT,
                message="Transaction has timed out",
                clear=True,  # signal to clear the active transaction
            )
        return pending.params


class FixedRateSpeedUp(AsyncTxStrategy):
    """Speedup strategy for pending transactions."""

    SPEEDUP_FACTOR = 1.125  # 12.5% increase
    MAX_TIP = Gwei(1)  # gwei maxPriorityFeePerGas per transaction

    _NAME = f"speedup-{SPEEDUP_FACTOR}%"

    def _calculate_speedup_fee(self, pending: TxParams) -> Tuple[Wei, Wei]:
        base_fee = self.w3.eth.get_block("latest")["baseFeePerGas"]
        suggested_tip = self.w3.eth.max_priority_fee
        _log_gas_weather(base_fee, suggested_tip)
        max_priority_fee = round(
            max(pending["maxPriorityFeePerGas"], suggested_tip) * self.SPEEDUP_FACTOR
        )
        max_fee_per_gas = round(
            max(
                pending["maxFeePerGas"] * self.SPEEDUP_FACTOR,
                (base_fee * 2) + max_priority_fee,
            )
        )
        return max_priority_fee, max_fee_per_gas

    def execute(self, pending: PendingTx) -> TxParams:
        params = pending.params
        old_tip, old_max_fee = params["maxPriorityFeePerGas"], params["maxFeePerGas"]
        new_tip, new_max_fee = self._calculate_speedup_fee(params)
        tip_increase = round(Web3.from_wei(new_tip - old_tip, "gwei"), 4)
        fee_increase = round(Web3.from_wei(new_max_fee - old_max_fee, "gwei"), 4)

        if new_tip > self.MAX_TIP:
            raise Wait(
                f"Pending transaction maxPriorityFeePerGas exceeds spending cap {self.MAX_TIP}"
            )

        latest_nonce = self.w3.eth.get_transaction_count(params["from"], "latest")
        pending_nonce = self.w3.eth.get_transaction_count(params["from"], "pending")
        if pending_nonce - latest_nonce > 0:
            log.warn("Overriding pending transaction!")

        log.info(
            f"Speeding up transaction #{params['nonce']} \n"
            f"maxPriorityFeePerGas (~+{tip_increase} gwei) {old_tip} -> {new_tip} \n"
            f"maxFeePerGas (~+{fee_increase} gwei) {old_max_fee} -> {new_max_fee}"
        )
        params = dict(params)
        params["maxPriorityFeePerGas"] = new_tip
        params["maxFeePerGas"] = new_max_fee
        params["nonce"] = latest_nonce
        params = TxParams(params)
        return params



import sys

from twisted.logger import Logger, textFileLogObserver, globalLogPublisher

log = Logger("AutomaticTxMachine")


observer = textFileLogObserver(sys.stdout)

globalLogPublisher.addObserver(observer)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Dict, Optional

from eth_typing import ChecksumAddress
from eth_utils import encode_hex
from hexbytes import HexBytes
from web3.types import TxData, TxParams, TxReceipt

from atxm.exceptions import Faults

TxHash = HexBytes


@dataclass
class AsyncTx(ABC):
    id: int
    final: bool = field(default=None, init=False)
    fault: Optional[Faults] = field(default=None, init=False)
    on_broadcast: Optional[Callable] = field(default=None, init=False)
    on_finalized: Optional[Callable] = field(default=None, init=False)
    on_pause: Optional[Callable] = field(default=None, init=False)
    on_fault: Optional[Callable] = field(default=None, init=False)

    def __repr__(self):
        return f"<{self.__class__.__name__} id={self.id} final={self.final}>"

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    @abstractmethod
    def to_dict(self):
        raise NotImplementedError

    @classmethod
    def from_dict(cls, data: Dict):
        raise NotImplementedError


@dataclass
class FutureTx(AsyncTx):
    final: bool = field(default=False, init=False)
    params: TxParams
    _from: ChecksumAddress
    info: Optional[Dict] = None

    def __hash__(self):
        return hash(self.id)

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "from": self._from,
            "params": _serialize_tx_params(self.params),
            "info": self.info,
        }

    @classmethod
    def from_dict(cls, data: Dict):
        return cls(
            id=int(data["id"]),
            _from=data["from"],
            params=TxParams(data["params"]),
            info=dict(data["info"]),
        )


@dataclass
class PendingTx(AsyncTx):
    final: bool = field(default=False, init=False)
    txhash: TxHash
    created: int
    data: Optional[TxData] = None

    def __hash__(self) -> int:
        return hash(self.txhash)

    def to_dict(self) -> Dict:
        return {
            "id": self.id,
            "txhash": self.txhash.hex(),
            "created": self.created,
            "data": self.data,
        }

    @classmethod
    def from_dict(cls, data: Dict):
        return cls(
            id=int(data["id"]),
            txhash=HexBytes(data["txhash"]),
            created=int(data["created"]),
            data=dict(data) if data else dict(),
        )


@dataclass
class FinalizedTx(AsyncTx):
    final: bool = field(default=True, init=False)
    receipt: TxReceipt

    def __hash__(self) -> int:
        return hash(self.receipt["transactionHash"])

    def to_dict(self) -> Dict:
        return {"id": self.id, "receipt": _serialize_tx_receipt(self.receipt)}

    @classmethod
    def from_dict(cls, data: Dict):
        receipt = _deserialize_tx_receipt(data["receipt"])
        return cls(id=int(data["id"]), receipt=receipt)


@dataclass
class FaultyTx(AsyncTx):
    final: bool = field(default=False, init=False)
    fault: Faults
    error: Optional[str] = None

    def __hash__(self) -> int:
        return hash(self.id)

    def to_dict(self) -> Dict:
        return {"id": self.id, "error": str(self.error), "fault": self.fault.value}

    @classmethod
    def from_dict(cls, data: Dict):
        return cls(
            id=int(data["id"]), error=str(data["error"]), fault=Faults(data["fault"])
        )


def _serialize_tx_params(params: TxParams) -> Dict:
    """Serializes transaction parameters to a dictionary for disk storage."""
    data = params.get("data")
    if isinstance(data, bytes):
        data = encode_hex(data)
    result = {
        "nonce": params["nonce"],
        "chainId": params["chainId"],
        "gas": params["gas"],
        "type": params.get("type"),
        "to": params["to"],
        "value": params["value"],
        "data": data,
        "maxPriorityFeePerGas": params.get("maxPriorityFeePerGas"),
        "maxFeePerGas": params.get("maxFeePerGas"),
        "gasPrice": params.get("gasPrice"),
    }
    # filtrate Nones
    result = {k: v for k, v in result.items() if v is not None}
    return result


def _serialize_tx_receipt(receipt: TxReceipt) -> Dict:
    return {
        "transactionHash": encode_hex(receipt["transactionHash"]),
        "transactionIndex": receipt["transactionIndex"],
        "blockHash": encode_hex(receipt["blockHash"]),
        "blockNumber": receipt["blockNumber"],
        "from": receipt["from"],
        "to": receipt["to"],
        "cumulativeGasUsed": receipt["cumulativeGasUsed"],
        "gasUsed": receipt["gasUsed"],
        "status": receipt["status"],
    }


def _deserialize_tx_receipt(receipt: Dict) -> TxReceipt:
    return TxReceipt(
        {
            "transactionHash": HexBytes(receipt["transactionHash"]),
            "transactionIndex": receipt["transactionIndex"],
            "blockHash": HexBytes(receipt["blockHash"]),
            "blockNumber": receipt["blockNumber"],
            "from": receipt["from"],
            "to": receipt["to"],
            "cumulativeGasUsed": receipt["cumulativeGasUsed"],
            "gasUsed": receipt["gasUsed"],
            "status": receipt["status"],
        }
    )



from copy import copy
from typing import List, Set

from eth_account.signers.local import LocalAccount
from web3.types import TxParams

from atxm.machine import _Machine
from atxm.tx import (
    FinalizedTx,
    FutureTx,
    PendingTx,
    AsyncTx,
)


class AutomaticTxMachine(_Machine):
    def start(self, now: bool = False) -> None:
        """Start the machine. if now is True, start immediately."""
        super()._start(now=now)

    def stop(self) -> None:
        """Stop the machine."""
        super()._stop()

    @property
    def running(self) -> bool:
        """Return True if the machine is running."""
        return bool(self._task.running)

    @property
    def paused(self) -> bool:
        """Return True if the machine is paused."""
        return bool(self.__pause)

    @property
    def busy(self) -> bool:
        """Returns True if the machine is busy."""
        return super()._busy

    @property
    def queued(self) -> List[FutureTx]:
        """Return a list of queued transactions."""
        return list(self._state.queue)

    @property
    def pending(self) -> PendingTx:
        """Return the active transaction if there is one."""
        return copy(self._state.pending or None)

    @property
    def finalized(self) -> Set[FinalizedTx]:
        """Return a set of finalized transactions."""
        return set(self._state.finalized)

    @property
    def faults(self) -> List[AsyncTx]:
        """Return a set of faulted transactions."""
        return list(self._state.faulty)

    def queue_transaction(
        self, params: TxParams, signer: LocalAccount, *args, **kwargs
    ) -> FutureTx:
        """
        Queue a new transaction for broadcast and subsequent tracking.
        Optionally provide a dictionary of additional string data
        to log during the transaction's lifecycle for identification.
        """
        if signer.address not in self.signers:
            self.signers[signer.address] = signer
        tx = self._state._queue(_from=signer.address, params=params, *args, **kwargs)
        if not self._task.running:
            self._wake()
        return tx

    def queue_transactions(
        self, params: List[TxParams], signer: LocalAccount, *args, **kwargs
    ) -> List[FutureTx]:
        """
        Queue a list of transactions for broadcast and subsequent tracking.

        Sorts incoming transactions by nonce. The machine is tolerant
        to nonce collisions, but it's best to avoid them when possible,
        plus it's a good practice to broadcast transactions in the
        order they were originally created in by the caller.
        """
        params = sorted(params, key=lambda x: x["nonce"])

        future_txs = []
        for _params in params:
            future_txs.append(
                self.queue_transaction(signer=signer, params=_params, *args, **kwargs)
            )
        return future_txs



import json
import time
from collections import deque
from copy import copy
from json import JSONDecodeError
from pathlib import Path
from typing import Callable, Deque, Dict, Optional, Set, Tuple

from eth_typing import ChecksumAddress
from web3.types import TxParams, TxReceipt

from atxm.exceptions import Faults
from atxm.logging import log
from atxm.tx import (
    FinalizedTx,
    FutureTx,
    PendingTx,
    TxHash,
    AsyncTx,
    FaultyTx,
)
from atxm.utils import fire_hook


class _State:
    """State management for transaction tracking."""

    _FILEPATH = ".txs.json"
    __COUNTER = 0  # id generator
    __FAULT_CACHE_SIZE = 10

    def __init__(self, disk_cache: bool, filepath: Optional[Path] = None):
        self.__filepath = filepath or self._FILEPATH

        self.__queue: Deque[FutureTx] = deque()
        self.__active: Optional[PendingTx] = None
        self.finalized: Set[FinalizedTx] = set()
        self.faulty: Deque[AsyncTx] = deque(maxlen=self.__FAULT_CACHE_SIZE)

        self.disk_cache = disk_cache
        if disk_cache:
            self.restore()

    def to_dict(self) -> Dict:
        """Serialize the state to a JSON string."""
        active = self.__active.to_dict() if self.__active else {}
        queue = [tx.to_dict() for tx in self.__queue]
        finalized = [tx.to_dict() for tx in self.finalized]
        _dict = {"queue": queue, "active": active, "finalized": finalized}
        return _dict

    def commit(self) -> None:
        """Write the state to the cache file."""
        if not self.disk_cache:
            return
        with open(self.__filepath, "w+t") as file:
            data = json.dumps(self.to_dict())
            file.write(data)
        log.debug(f"[state] wrote transaction cache file {self.__filepath}")

    def restore(self) -> bool:
        """
        Restore the state from the cache file.
        Returns True if the cache file exists and was successfully
        restored with data.
        """
        if not self.disk_cache:
            return False

        # read & parse
        if not self.__filepath.exists():
            return False
        with open(self.__filepath, "r+t") as file:
            data = file.read()
        try:
            data = json.loads(data)
        except JSONDecodeError:
            data = dict()
        active = data.get("active", dict())
        queue = data.get("queue", list())
        final = data.get("finalized", list())

        # deserialize & restore
        self.__active = PendingTx.from_dict(active) if active else None
        self.__queue.extend(FutureTx.from_dict(tx) for tx in queue)
        self.finalized = {FinalizedTx.from_dict(tx) for tx in final}
        log.debug(
            f"[state] restored {len(queue)} transactions from cache file {self.__filepath}"
        )

        return bool(data)

    def __set_active(self, tx: PendingTx) -> None:
        """Update the active transaction (destructive operation)."""
        old = None
        if self.__active:
            old = self.__active.txhash
        self.__active = tx
        self.commit()
        if old:
            log.debug(
                f"[state] updated active transaction {old.hex()} -> {tx.txhash.hex()}"
            )
            return
        log.debug(f"[state] tracked active transaction {tx.txhash.hex()}")

    def morph(self, tx: FutureTx, txhash: TxHash) -> PendingTx:
        """
        Morphs a future transaction into a pending transaction.
        Uses polymorphism to transform the future transaction into a pending transaction.
        """
        tx.txhash = txhash
        tx.created = int(time.time())
        tx.capped = False
        tx.__class__ = PendingTx
        tx: PendingTx
        self.__set_active(tx=tx)
        return tx

    def fault(
        self,
        fault: Faults,
        clear_active: bool,
        error: Optional[str] = None,
    ) -> None:
        """Fault the active transaction."""
        hook = self.__active.on_fault
        if not self.__active:
            raise RuntimeError("No active transaction to fault")
        tx = self.__active
        tx.fault = fault
        tx.error = error
        tx.__class__ = FaultyTx
        tx: FaultyTx
        self.faulty.append(tx)
        log.warn(
            f"[state] tracked faulty transaction #atx-{tx.id} "
            f"{len(self.faulty)}/{self.faulty.maxlen} faults cached"
        )
        if clear_active:
            self.clear_active()
        if hook:
            fire_hook(hook=hook, tx=tx)

    def finalize_active_tx(self, receipt: TxReceipt) -> None:
        """
        Finalizes a pending transaction.
        Use polymorphism to transform the pending transaction into a finalized transaction.
        """
        hook = self.__active.on_finalized
        if not self.__active:
            raise RuntimeError("No pending transaction to finalize")
        self.__active.receipt = receipt
        self.__active.__class__ = FinalizedTx
        tx = self.__active
        self.finalized.add(tx)  # noqa
        log.info(f"[state] #atx-{tx.id} pending -> finalized")
        self.clear_active()
        if hook:
            fire_hook(hook=hook, tx=tx)

    def clear_active(self) -> None:
        """Clear the active transaction (destructive operation)."""
        self.__active = None
        self.commit()
        log.debug(
            f"[state] cleared 1 pending transaction \n"
            f"[state] {len(self.queue)} queued "
            f"transaction{'s' if len(self.queue) != 1 else ''} remaining"
        )

    @property
    def pending(self) -> Optional[PendingTx]:
        """Return the active pending transaction if there is one."""
        return copy(self.__active)

    @property
    def queue(self) -> Tuple[FutureTx, ...]:
        """Return the queue of transactions."""
        return tuple(self.__queue)

    def _pop(self) -> FutureTx:
        """Pop the next transaction from the queue."""
        return self.__queue.popleft()

    def _requeue(self, tx: FutureTx) -> None:
        """Re-queue a transaction for broadcast and subsequent tracking."""
        self.__queue.append(tx)
        self.commit()
        log.info(
            f"[state] re-queued transaction #atx-{tx.id} "
            f"priority {len(self.__queue)}"
        )

    def _queue(
        self,
        params: TxParams,
        _from: ChecksumAddress,
        info: Dict[str, str] = None,
        on_broadcast: Optional[Callable] = None,
        on_finalized: Optional[Callable] = None,
        on_pause: Optional[Callable] = None,
        on_fault: Optional[Callable] = None,
    ) -> FutureTx:
        """Queue a new transaction for broadcast and subsequent tracking."""
        tx = FutureTx(
            _from=_from,
            id=self.__COUNTER,
            params=params,
            info=info,
        )

        # configure hooks
        tx.on_broadcast = on_broadcast
        tx.on_finalized = on_finalized
        tx.on_pause = on_pause
        tx.on_fault = on_fault

        self.__queue.append(tx)
        self.commit()
        self.__COUNTER += 1
        log.info(
            f"[state] queued transaction #atx-{tx.id} " f"priority {len(self.__queue)}"
        )
        return tx
